# -*- coding:utf-8 -*-
import numpy as np
import pandas as pd
import seaborn as sn
from matplotlib import pyplot as plt
from sklearn.metrics import classification_report, confusion_matrix, cohen_kappa_score

from TorchMiner import BasePlugin
from TorchMiner.utils import figure2numpy


class MultiClassesClassificationMetric(BasePlugin):
    """MultiClassesClassificationMetric
    This can be used directly if your loss function is torch.nn.CrossEntropy
    """
    requirements = ["TensorboardDrawer"]

    def __init__(
            self,
            accuracy=True,
            confusion=True,
            kappa_score=True,
            report=True,
            backend="TensorboardDrawer",
            forward=None,
            labels=None
    ):
        super().__init__()
        self.accuracy = accuracy
        self.confusion_matrix = confusion
        self.kappa_score = kappa_score
        self.classification_report = report
        self.backend = backend  # TODO:Can be used for sheet recorder
        self.forward = forward
        self.labels = labels

    # def before_init(self):
    #     self.create_sheet_column("latest_confusion_matrix", "Latest Confusion Matrix")
    #     self.create_sheet_column("kappa_score", "Kappa Score")
    #     self.create_sheet_column("accuracy", "Accuracy")

    def after_init(self, *args, **kwargs):
        self.recorder = self.miner.plugins.get(self.backend)

    def before_val_epoch_start(self, epoch, **ignore):
        # TODO:Recommended to attach these attributes to Miner Obj, for combination with Other Plugins
        self.predicts = []
        self.label = []

    def predicts_and_labels(self, predicts, data):
        """
        Inherit this function to use custom classification function.
        :param predicts:
        :param data:
        :return:
        """
        predict = np.argmax(predicts, axis=1)  # Batch first
        label = data[1].cpu().numpy()  # label
        return predict, label

    def after_val_iteration_ended(self, predicts, data, **ignore):
        raw_output = predicts.detach().cpu().numpy()
        predict, label = self.predicts_and_labels(raw_output, data)
        self.predicts.append(predict)  # fix #17
        self.label.append(label)

    def after_val_epoch_end(self, val_loss, **ignore):
        try:
            predicts = np.concatenate(self.predicts)
            label = np.concatenate(self.label)
        except Exception as e:  # Fix #13
            self.logger.warning(f"{e} when concatenating val predictions, maybe only have one batch of val iteration.")
            predicts = self.predicts[0]
            label = self.label[0]

        if self.accuracy:
            accuracy = np.array(predicts == label).sum() / len(predicts)
            self.logger.info(f"Val Accuracy:{accuracy}")
            self.recorder.scalar("Val/Accuracy", accuracy)

        if self.confusion_matrix:
            matrix = confusion_matrix(label, predicts, labels=self.labels)  # Now can pass string items
            df_cm = pd.DataFrame(matrix)
            svm = sn.heatmap(df_cm, annot=True, cmap="OrRd", fmt=".3g")
            figure = svm.get_figure()
            if val_loss < self.miner.lowest_val_loss:
                self.recorder.figure("Val/ConfusionMatrix", figure2numpy(figure))
            plt.close(figure)

        if self.kappa_score:
            kappa = cohen_kappa_score(label, predicts, weights="quadratic")
            self.recorder.scalar("Val/KappaScore", kappa)

        if self.classification_report:
            # TODO:Design a better way to output or record classification report
            print(classification_report(label, predicts))



# -*- coding:utf-8 -*-
"""
Global Logger
"""
import logging

# https://stackoverflow.com/questions/384076/how-can-i-color-python-logging-output
FORMAT = "%(asctime)s - [$BOLD%(name)-s$RESET][%(levelname)-s]($BOLD%(filename)s$RESET;$BOLD%(funcName)s$RESET:%(lineno)d): %(message)s "
# FORMAT = "%(message)s "

BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)
# The background is set with 40 plus the number of the color, and the foreground with 30
# These are the sequences need to get colored ouput
RESET_SEQ = "\033[0m"
COLOR_SEQ = "\033[1;%dm"
BOLD_SEQ = "\033[1m"


def formatter_message(message, use_color=True):
    if use_color:
        message = message.replace("$RESET", RESET_SEQ).replace("$BOLD", BOLD_SEQ)
    else:
        message = message.replace("$RESET", "").replace("$BOLD", "")
    return message


COLORS = {
    'WARNING': YELLOW,
    'INFO': WHITE,
    'DEBUG': BLUE,
    'CRITICAL': YELLOW,
    'ERROR': RED
}


class ColoredFormatter(logging.Formatter):
    def __init__(self, msg, use_color=True):
        logging.Formatter.__init__(self, msg)
        self.use_color = use_color

    def format(self, record):
        levelname = record.levelname
        if self.use_color and levelname in COLORS:
            levelname_color = COLOR_SEQ % (30 + COLORS[levelname]) + levelname + RESET_SEQ
            record.levelname = levelname_color
        return logging.Formatter.format(self, record)


# Custom logger class with multiple destinations
class ColoredLogger(logging.Logger):
    COLOR_FORMAT = formatter_message(FORMAT, True)

    def __init__(self, name):
        logging.Logger.__init__(self, name, logging.DEBUG)
        color_formatter = ColoredFormatter(self.COLOR_FORMAT)
        console = logging.StreamHandler()
        console.setFormatter(color_formatter)
        self.addHandler(console)
        return


def getLogger(name):
    logging.setLoggerClass(ColoredLogger)
    return logging.getLogger(name)



import io

import numpy as np


def seed_everything(seed):
    """
    Fix the seed for generating random numbers.
    :param seed:
    :return:
    """
    import torch
    from numpy import random

    torch.manual_seed(seed)
    # may lead to bad performance
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    random.seed(seed)

    if seed == 0:
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False


def create_dir(directory, *args):
    """Create directory"""
    import os

    current_dir = directory
    for dir_name in args:
        current_dir = os.path.join(current_dir, dir_name)
    if not os.path.isdir(current_dir):
        os.mkdir(current_dir)


def figure2numpy(fig):
    # Solution to store Matplotlib Figure in TensorBoardX
    # https://stackoverflow.com/questions/7821518/matplotlib-save-plot-to-numpy-array
    with io.BytesIO() as buff:
        fig.savefig(buff, format='raw')
        buff.seek(0)
        data = np.frombuffer(buff.getvalue(), dtype=np.uint8)
    w, h = fig.canvas.get_width_height()
    im = data.reshape((int(h), int(w), -1))
    im = im.transpose((2, 0, 1))  # Change to CHW
    return im



# -*- coding:utf-8 -*-
import math

import httplib2
from google.oauth2 import service_account
from googleapiclient.discovery import build
from googleapiclient.errors import HttpError
from googleapiclient.http import MediaFileUpload

from TorchMiner.plugins.Recorder.BasicSheet import _async, BaseSheet


def num_to_letter(num):
    num += 1
    letters = ""
    while num:
        mod = (num - 1) % 26
        letters += chr(mod + 65)
        num = (num - 1) // 26
    return "".join(reversed(letters))


class GoogleSheet(BaseSheet):
    def _create_experiment_row(self):
        pass

    def __init__(
            self, sheet_id, service_account_file, meta_prefix="", build_kwargs=None, proxy=None
    ):
        super().__init__()
        if proxy:  # If proxy is Used to connect to Google Api Server
            # Proxy Should be a dict :{"ip":"ip_address","port":"your port"}
            self.http = httplib2.Http(proxy_info=httplib2.ProxyInfo(
                httplib2.socks.PROXY_TYPE_HTTP, proxy["ip"], proxy["port"]
            ))
        else:
            self.http = None

        if build_kwargs is None:
            build_kwargs = {}

        self.sheet_id = sheet_id
        self.meta_prefix = meta_prefix

        credentials = service_account.Credentials.from_service_account_file(
            service_account_file, http=self.http
        )
        service = build("sheets", "v4", credentials=credentials, **build_kwargs, http=self.http)
        self.drive = build("drive", "v3", credentials=credentials, http=self.http)
        self.sheet = service.spreadsheets()
        self.drive_folder_id = self._prepare_drive_directory()
        self.sheet_progress = {
            "epoch": "0",
            "train_percentage": "0%",
            "val_percentage": "0%",
        }

    def _index_of(self, key):
        search = {
            "dataFilters": [
                {"developerMetadataLookup": {"metadataKey": key}}
            ]
        }
        result = (
            self.sheet.developerMetadata()
                .search(spreadsheetId=self.sheet_id, body=search)
                .execute()
        )
        if len(result.items()) == 0:
            return False
        else:
            return result["matchedDeveloperMetadata"][0]["developerMetadata"]["location"]["dimensionRange"][
                "startIndex"]

    _exists = _index_of  # alias

    @_async
    def reset_index(self):
        self.banner_index = self._create_banner_dimension()
        self.title_index = self._create_title_dimension()
        self.endcol_index = self._create_end_column_divider()
        self.experiment_row_index = self._insert_dimension(
            self.experiment_row_name, self.title_index + 1, "ROWS"
        )

    @property
    def dark_bg(self):
        return {
            "red": 0.10980392156862745,
            "green": 0.5686274509803921,
            "blue": 0.6039215686274509,
        }

    @property
    def white(self):
        return {"red": 1.0, "green": 1.0, "blue": 1.0}

    @property
    def light_bg(self):
        return {
            "red": 0.9411764705882353,
            "green": 1.0,
            "blue": 0.9882352941176471,
        }

    def _create_banner_dimension(self):
        return self._insert_dimension(self.banner_row_name, 0, "ROWS")

    def _create_title_dimension(self):
        return self._insert_dimension(self.title_row_name, 1, "ROWS")

    def _create_end_column_divider(self):
        icol = self._insert_dimension(self.end_column_name, 0, "COLUMNS")
        requests = [
            {
                "repeatCell": {
                    "range": {
                        "sheetId": 0,
                        "startColumnIndex": icol,
                        "endColumnIndex": icol + 1,
                        "startRowIndex": self.title_index,
                        "endRowIndex": self.title_index + 1,
                    },
                    "cell": {
                        "userEnteredFormat": {
                            "backgroundColor": self.dark_bg,
                            "textFormat": {
                                "foregroundColor": self.white,
                                "bold": False,
                                "fontSize": 12,
                            },
                            "horizontalAlignment": "CENTER",
                            "verticalAlignment": "MIDDLE",
                        }
                    },
                    "fields": "*",
                }
            },
            {
                "updateDimensionProperties": {
                    "range": {
                        "sheetId": 0,
                        "dimension": "COLUMNS",
                        "startIndex": icol,
                        "endIndex": icol + 1,
                    },
                    "properties": {
                        "hiddenByUser": True,
                    },
                    "fields": "hiddenByUser",
                }
            },
        ]

        self.sheet.batchUpdate(
            spreadsheetId=self.sheet_id, body={"requests": requests}
        ).execute()
        return icol

    def _insert_dimension(self, row_name, index, dim, extra_request=None):
        result = self._exists(row_name)
        if result is not False:
            return result

        if extra_request is None:
            extra_request = []

        create_row_request = {
            "insertDimension": {
                "range": {
                    "sheetId": 0,
                    "dimension": dim,
                    "startIndex": index,
                    "endIndex": index + 1,
                },
                "inheritFromBefore": False,
            }
        }

        assign_name_request = {
            "createDeveloperMetadata": {
                "developerMetadata": {
                    "metadataKey": row_name,
                    "metadataValue": row_name,
                    "location": {
                        "dimensionRange": {
                            "sheetId": 0,
                            "dimension": dim,
                            "startIndex": index,
                            "endIndex": index + 1,
                        }
                    },
                    "visibility": "DOCUMENT",
                }
            }
        }

        body = {
            "requests": [create_row_request, assign_name_request, *extra_request],
        }

        self.sheet.batchUpdate(spreadsheetId=self.sheet_id, body=body).execute()
        return self._exists(row_name)

    @_async
    def onready(self):
        banner = """
        TorchMiner Official Google Sheet Plugin.
        If you found it's useful, please considering star the project at https://github.com/InEase/TorchMiner.
        """
        icol_end = self._index_of(self.end_column_name)
        icol_start = self._index_of(self.end_column_name) - len(self.columns)
        merge_cells = {
            "mergeCells": {
                "range": {
                    "sheetId": 0,
                    "startRowIndex": self.banner_index,
                    "endRowIndex": self.banner_index + 1,
                    "startColumnIndex": icol_start,
                    "endColumnIndex": icol_end,
                },
                "mergeType": "MERGE_ALL",
            }
        }

        change_cell = {
            "repeatCell": {
                "range": {
                    "sheetId": 0,
                    "startRowIndex": self.banner_index,
                    "endRowIndex": self.banner_index + 1,
                    "startColumnIndex": icol_start,
                    "endColumnIndex": icol_end,
                },
                "cell": {"userEnteredFormat": {"wrapStrategy": "WRAP"}},
                "fields": "*",
            }
        }

        auto_resize = {
            "autoResizeDimensions": {
                "dimensions": {
                    "sheetId": 0,
                    "dimension": "ROWS",
                    "startIndex": self.banner_index,
                    "endIndex": self.banner_index + 1,
                }
            }
        }

        body = {"requests": [merge_cells, change_cell, auto_resize]}

        self.sheet.batchUpdate(spreadsheetId=self.sheet_id, body=body).execute()
        self._update_cells(
            f"{num_to_letter(icol_start)}{self.banner_index + 1}", [banner]
        )

    @_async
    def create_column(self, key, title, size=None):
        super().create_column(key, title)
        col_index = self._index_of(self.end_column_name)

        change_cell_request = {
            "repeatCell": {
                "range": {
                    "sheetId": 0,
                    "startColumnIndex": col_index,
                    "endColumnIndex": col_index + 1,
                    "startRowIndex": self.experiment_row_index,
                    "endRowIndex": self.experiment_row_index + 1,
                },
                "cell": {
                    "userEnteredFormat": {
                        "backgroundColor": self.light_bg,
                        "horizontalAlignment": "CENTER",
                        "verticalAlignment": "MIDDLE",
                    }
                },
                "fields": "*",
            },
        }

        assign_name_request = {
            "createDeveloperMetadata": {
                "developerMetadata": {
                    "metadataKey": "__minetorch_column__",
                    "metadataValue": key,
                    "location": {
                        "dimensionRange": {
                            "sheetId": 0,
                            "dimension": "COLUMNS",
                            "startIndex": col_index,
                            "endIndex": col_index + 1,
                        }
                    },
                    "visibility": "DOCUMENT",
                }
            }
        }

        index = self._insert_dimension(
            key, col_index, "COLUMNS", [change_cell_request, assign_name_request]
        )
        self._update_cells(f"{num_to_letter(index)}{self.title_index + 1}", [title])

    def _update_cells(self, a1, values):
        value_range = {"range": a1, "majorDimension": "ROWS", "values": [values]}
        try:
            self.sheet.values().update(
                spreadsheetId=self.sheet_id,
                range=a1,
                valueInputOption="USER_ENTERED",
                body=value_range,
            ).execute()
        except Exception as e:
            self.logger.warn(f"Update sheet failed with {e}")
            return

    @_async
    def flush(self):
        self.logger.debug("GoogleSheet Flushing...")
        irow = self._index_of(self.experiment_row_name)
        column_indices = self._get_column_indices()
        for key, value in self.cached_row_data.items():
            raw_value = value.get("raw")
            processor = value.get("processor")
            if processor is None:
                value = raw_value
            else:
                value = getattr(self, f"_process_{processor}")(key, raw_value)
            icol = column_indices[key]
            self._update_cells(f"{num_to_letter(icol)}{irow + 1}", [value])
        self.cached_row_data = {}

    def update_progress(self, **kwargs):
        self.sheet_progress.update(kwargs)
        # TODO:This Progress should not be divided into two parts
        progress = f"""
                    epoch:  {self.sheet_progress["epoch"]}
                    train progress:  {self.sheet_progress["train_percentage"]}
                    val progress:  {self.sheet_progress["val_percentage"]}
                    """
        self.update("progress", progress)
        # TODO:This flush method seems to have problems, if flush failed ,information got lost
        self.periodly_flush()

    def _process_upload_image(self, key, value, retry=True):
        try:
            image_id = self._upload_drive_image(key, value)
            return f'=IMAGE("https://drive.google.com/uc?export=view&id={image_id}", 2)'
        except HttpError as e:
            if retry:
                return self._process_upload_image(key, value)
            raise e

    # def _process_repr(self, key, value):
    #     return repr(value)

    def _get_column_indices(self):
        search = {
            "dataFilters": [
                {
                    "developerMetadataLookup": {
                        "metadataKey": "__minetorch_column__"
                    }
                }
            ]
        }
        r = (
            self.sheet.developerMetadata()
                .search(spreadsheetId=self.sheet_id, body=search)
                .execute()
        )
        result = {}
        for item in r["matchedDeveloperMetadata"]:
            column_key = item["developerMetadata"]["metadataValue"]
            index = item["developerMetadata"]["location"]["dimensionRange"][
                "startIndex"
            ]
            result[column_key] = index
        return result

    def _upload_drive_image(self, key, value, retry=True):
        try:
            file_metadata = {"name": key, "parents": [self.drive_folder_id]}
            media = MediaFileUpload(value, mimetype="image/png")
            file = (
                self.drive.files()
                    .create(body=file_metadata, media_body=media, fields="id")
                    .execute()
            )
            return file.get("id")
        except HttpError as e:
            if not retry:
                raise e
            self.drive_folder_id = self._prepare_drive_directory()
            self._upload_drive_image(key, value, retry=False)

    def _prepare_drive_directory(self):
        try:
            result = (
                self.drive.files()
                    .list(q="name='TorchMiner_assets'", fields="files(id)")
                    .execute()
            )
            dir_id = result["files"][0]["id"]
        except IndexError:
            file_metadata = {
                "name": "TorchMiner_assets",
                "mimeType": "application/vnd.google-apps.folder",
            }
            file = self.drive.files().create(body=file_metadata, fields="id").execute()
            dir_id = file.get("id")

        self.drive.permissions().create(
            fileId=dir_id, body={"role": "writer", "type": "anyone"}
        ).execute()
        return dir_id

    # Define Hooks
    def prepare(self, miner, *args, **kwargs):
        super(GoogleSheet, self).prepare(miner)
        self.set_miner(self)
        self.reset_index()
        self.create_column("code", "Code")
        self.create_column("progress", "Progress")
        self.create_column("loss", "Loss")
        self.update("code", miner.experiment)

    def before_init(self, *args, **kwargs):
        self.last_flushed_at = 0
        self.onready()
        self.flush()

    def before_epoch_start(self, epoch, *args, **kwargs):
        self.update_progress(epoch=epoch, train_percentage="0%", val_percentage="0%")

    def after_epoch_end(self):
        self.flush()

    def after_train_iteration_end(self, index, train_iters, *args, **kwargs):
        self.update_progress(train_percentage=math.ceil(index / train_iters * 100))

    def before_val_iteration_start(self, index, val_iters, *args, **kwargs):
        self.update_progress(val_percentage=math.ceil(index / val_iters * 100))



import torch
from minetorch.plugin import Plugin
from minetorch.statable import Statable


class NoiseSampleDetector(Plugin, Statable):
    """This plugin helps to find out the suspicious noise samples.
    provid a metric which compute a scalar for every sample, in most cases
    the metric should be the loss function without reduce.
    """

    def __init__(self, metric, topn=50):
        super().__init__()
        self.metric = metric
        self.topn = topn
        self.train_metrics = []
        self.val_metrics = []

    def before_init(self):
        self.miner.statable[self.__class__.__name__] = self
        self.train_dataloader = torch.utils.data.DataLoader(
            self.miner.train_dataloader.dataset,
            batch_size=self.miner.train_dataloader.batch_size,
            num_workers=self.miner.train_dataloader.num_workers,
            shuffle=False,
        )
        self.val_dataloader = torch.utils.data.DataLoader(
            self.miner.val_dataloader.dataset,
            batch_size=self.miner.train_dataloader.batch_size,
            num_workers=self.miner.train_dataloader.num_workers,
            shuffle=False,
        )

    def load_state_dict(self, data):
        self.train_metrics = data[0]
        self.val_metrics = data[1]

    def state_dict(self):
        return (self.train_metrics, self.val_metrics)

    def after_epoch_end(self, **kwargs):
        with torch.no_grad():
            self.train_metrics.append(self._predict_dataset(self.train_dataloader))
            self.val_metrics.append(self._predict_dataset(self.val_dataloader))

        _, train_indices = torch.sort(
            torch.std(torch.stack(self.train_metrics), dim=0), descending=True
        )
        _, val_indices = torch.sort(
            torch.std(torch.stack(self.val_metrics), dim=0), descending=True
        )

        self.print_txt(
            f"Train dataset most {self.topn} suspicious indices: {train_indices.tolist()[:self.topn]} \n"
            f"Validation dataset most {self.topn} suspicious indices: {val_indices.tolist()[:self.topn]}",
            "suspicious_noise_samples",
        )

    def _predict_dataset(self, dataloader):
        results = torch.zeros([len(dataloader.dataset)])
        for index, data in enumerate(dataloader):
            predict = self.model(data[0].to(self.devices))
            offset = index * dataloader.batch_size
            results[offset: offset + dataloader.batch_size] = (
                self.metric(predict, data[1].to(self.devices)).detach().cpu()
            )
        return results



# -*- coding:utf-8 -*-
import numpy as np
from sklearn.metrics import mean_absolute_percentage_error

from TorchMiner import BasePlugin


class RegressionAccuracy(BasePlugin):
    def before_train_epoch_start(self, *args, **kwargs):
        self.predicts = []
        self.label = []

    @staticmethod
    def forward(predicts, original_data):
        raw_output = predicts.detach().cpu().numpy()
        # predicts = np.argmax(raw_output, axis=1)  # Batch first
        label = original_data[1].cpu().numpy()  # label
        return raw_output, label  # 1 dim ndarray

    def after_val_iteration_ended(self, predicts, data, *args, **kwargs):
        data = self.forward(predicts, data)
        self.predicts.append(data[0])
        self.label.append(data[1])

    def after_epoch_end(self, *args, **kwargs):
        self.predicts = np.concatenate(self.predicts)
        self.label = np.concatenate(self.label)
        acc = mean_absolute_percentage_error(self.label, self.predicts)
        self.logger.info(f"Val Accuracy:{acc}")



class BasePlugin:
    requirements = []

    def __init__(self):
        self.name = self.__class__.__name__
        self.miner = None
        self.logger = None  # Plugin Logger will be inited in prepare stage

    def prepare(self, miner, *args, **kwargs):
        # Can be used for Monkey Patch Operations
        self.miner = miner

    # Plugin Data Begin
    def load_state_dict(self, state):
        pass

    def state_dict(self):
        return {}

    # Plugin Data End

    def set_miner(self, miner):
        """
        This function is deprecated.
        :param miner:
        :return:
        """
        self.miner = miner

    # Hook Functions Begin
    def before_init(self, *args, **kwargs):
        self.logger = self.miner.get_logger(self.name)

    def after_init(self, *args, **kwargs):
        pass

    def before_train_epoch_start(self, *args, **kwargs):
        pass

    def before_train_iteration_start(self, *args, **kwargs):
        pass

    def after_train_iteration_end(self, *args, **kwargs):
        pass

    def after_train_epoch_end(self, *args, **kwargs):
        pass

    def before_val_epoch_start(self, *args, **kwargs):
        pass

    def before_val_iteration_start(self, *args, **kwargs):
        pass

    def after_val_iteration_ended(self, *args, **kwargs):
        pass

    def after_val_epoch_end(self, *args, **kwargs):
        pass

    def before_checkpoint_persisted(self, *args, **kwargs):
        pass

    def after_checkpoint_persisted(self, *args, **kwargs):
        pass

    def before_quit(self, *args, **kwargs):
        pass


class PluginManager:
    def __init__(self, miner, plugins: list):
        self.miner = miner
        if plugins:
            self.plugins = plugins
            self.plugin_names = [i.name for i in self.plugins]
            self.maper = dict(zip(self.plugin_names, self.plugins))
            self.prepare()  # Prepare is the earliest
            self.logger = miner.get_logger("PluginManager")
            self.check_requirements()
        else:
            self.plugins = []

    def check_requirements(self):
        """
        Check Requirements of each Plugins
         - Requirements are set in restrict string
        :return:
        """
        error = False
        for p in self.plugins:
            self.logger.debug(f"Checking Requirements of {p.name}.")
            for r in p.requirements:
                if r not in self.plugin_names:
                    self.logger.error(f"Requirement {r} of {p.name} is not Find.")
                    error = True
        if not error:
            self.logger.info("Successfully Passed Plugin Requirements Check with no Errors.")
        # TODO: Try to import Unmet needs

    def status(self):
        """
        Print the Status of all registered Plugins
        :return:
        """
        self.logger.info(f"Registered Plugins:{self.plugins}")

    def call(self, hook, **payload):
        """
        Call Hook Functions
        :return:
        """
        for plugin in self.plugins:
            getattr(plugin, hook)(miner=self, **payload)  # !!! `miner=self` is totally different with just `self`

    def prepare(self):
        """
        prepare a given Plugin
        :return:
        """
        for p in self.plugins:
            p.prepare(self.miner)

    def load(self, checkpoint):
        # load plugin states
        for plugin in self.plugins:
            key = f"__plugin.{plugin.__class__.__name__}__"
            plugin.load_state_dict(checkpoint.get(key, {}))

    def save(self):
        temp = {}
        for p in self.plugins:
            temp[f"__plugin.{p.__class__.__name__}__"] = p.state_dict()
        return temp

    def get(self, name):
        """
        Get A Plugin From Plugin Manager.
        :param string name: Plugin name
        :return:
        """
        return self.maper[name]



import os
from pathlib import Path

import torch
import tqdm

from TorchMiner.Logger import ColoredLogger
from TorchMiner.plugins import PluginManager
from . import utils


class Miner(object):
    def __init__(
            self,
            alchemy_directory,
            model,
            optimizer,
            loss_func,
            experiment,
            train_dataloader=None,
            val_dataloader=None,
            resume=True,
            eval_epoch=1,
            persist_epoch=1,
            gpu=True,
            max_epochs=9999999,
            in_notebook=False,
            plugins=None,
            accumulated_iter=1,
            ignore_optimizer_resume=False,
            amp=False,
            amp_scaler=True,
    ):
        """
        Core Of TorchMiner
        :param alchemy_directory: The directory which TorchMiner will use to Store Everything in
        :param torch.nn.Module model: Target
        :param torch.optim.Optimizer optimizer:
        :param loss_func: A function to compute Loss
            A Special Function, the function receives 2 variable:
            * Miner: The Miner Object
            * Data: The Batch data yield by the loader
            return Value should be a float number of the loss.
        :param string experiment: Experiment Name
        :param torch.utils.data.DataLoader train_dataloader:

            --- Optional ---
        :param torch.utils.data.DataLoader val_dataloader: Default None. If None, skip Validation
        :param bool resume: Default True.
        :param int eval_epoch: Default 1. Validate every 'eval_epoch'
        :param int persist_epoch: Default 1. Save model every 'persist_epoch'
        :param gpu:
        :param plugins:
            The Differences between Hooks and Plugins:
                Hooks are Functions,They receive Miner and Payloads
                 - Each Hook can load One Function Each Time
                Plugins are Classes succeed to `TorchMiner.Plugin`
                 - Receives Miner when Plugin Inits
                 - Receives only payload when Hooks in Plugin was called
                 - One can use many plugins in a miner
        :param max_epochs:
        :param in_notebook:
        :param accumulated_iter:
        :param ignore_optimizer_resume:
        :param amp:
        :param amp_scaler:
        """
        self.alchemy_directory: Path = Path(alchemy_directory)  # working dir
        self.model = model
        self.train_dataloader = train_dataloader
        self.experiment: str = experiment
        self.logger_prototype = ColoredLogger
        self.val_dataloader = val_dataloader
        self.gpu = gpu
        self.in_notebook = in_notebook
        self.ignore_optimizer_resume = ignore_optimizer_resume
        self._create_dirs()
        self.devices = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.experiment_dir = alchemy_directory / self.experiment
        self.models_dir = alchemy_directory / self.experiment / "models"
        self.accumulated_iter = float(accumulated_iter)
        self.optimizer = optimizer
        self.loss_func = loss_func.to(self.devices)
        self.resume = resume
        self.eval_epoch = eval_epoch
        self.persist_stride = persist_epoch
        self.lowest_train_loss = float("inf")
        self.lowest_val_loss = float("inf")
        self.current_epoch = 0
        self.current_train_iteration = 0
        self.current_val_iteration = 0
        self.max_epochs = max_epochs
        self.amp = amp
        self.amp_scaler = amp_scaler
        if self.amp and self.amp_scaler:
            self.scaler = torch.cuda.amp.GradScaler()
        self.tqdm = tqdm.tqdm

        # --- Init Plugin ---
        self.plugins = PluginManager(self, plugins)
        self.logger = self.get_logger("Miner")
        # --- Before Init ---
        self.status = "init"
        self.plugins.call("before_init")
        self._init_model()
        # --- After Init ---
        self.plugins.call("after_init")

    @staticmethod
    def get_logger(name):
        return ColoredLogger(name)

    def _init_model(self):
        """resume from some checkpoint"""
        if isinstance(self.model, torch.nn.DataParallel):
            raise Exception(
                "Don't parallel the model yourself, instead, if the "
                "`gpu` option is true(default), TorchMiner will do this for you."
            )

        # TODO:简化Resume from pretrained的流程 添加对自定义路径的支持
        if self.resume is True:  # Find by TorchMiner
            # resume from the newest model
            if self._search_model_file("latest"):
                checkpoint_path = self._search_model_file("latest")
            else:
                checkpoint_path = None
                self.logger.warning("Could not find checkpoint to resume, " "train from scratch")
        elif isinstance(self.resume, str):  # specify model file name
            checkpoint_path = self._search_model_file(self.resume)
        elif isinstance(self.resume, int):  # specify train epoch
            checkpoint_path = self._search_model_file(self.resume)
        else:
            checkpoint_path = None

        if self.resume is not True and self.resume and checkpoint_path is None:
            # user has specified a none existed model, should raise a error
            raise Exception(f"Could not find model {self.resume}")

        if checkpoint_path is not None:
            # TODO:After Loading Checkpoint, output basic information
            self.logger.info(f"Start to load checkpoint {checkpoint_path}")
            checkpoint = torch.load(checkpoint_path)
            # Read Train Process From Resumed Data
            self.current_epoch = checkpoint.get("epoch", 0)
            self.current_train_iteration = checkpoint.get("train_iteration", 0)
            self.current_val_iteration = checkpoint.get("val_iteration", 0)
            self.lowest_train_loss = checkpoint.get("lowest_train_loss", 9999)
            self.lowest_val_loss = checkpoint.get("lowest_val_loss", 9999)

            # load model state
            try:
                self.model.load_state_dict(checkpoint["state_dict"], strict=True)
            except Exception as e:
                self.logger.warning(
                    f"load checkpoint failed with {e}, the state in the "
                    "checkpoint is not matched with the model, "
                    "try to reload checkpoint with unstrict mode"
                )
                # UnStrict Mode
                self.model.load_state_dict(checkpoint["state_dict"], strict=False)

            # load optimizer state
            if "optimizer" in checkpoint and not self.ignore_optimizer_resume:
                try:
                    self.optimizer.load_state_dict(checkpoint["optimizer"])
                except Exception as e:
                    self.logger.warning(
                        f"load optimizer state failed with {e}, will skip this error and continue, "
                        "stop the process if it is not expected"
                    )

            # load scaler state
            if self.amp and self.amp_scaler:
                try:
                    self.scaler.load_state_dict(checkpoint["scaler"])
                except Exception as e:
                    self.logger.warning(
                        f"load scaler state failed with {e}, will skip this error and continue, "
                        "stop the process if it is not expected"
                    )

            self.plugins.load(checkpoint)

            self.logger.info(f"Checkpoint {checkpoint_path} Successfully Loaded")
        # else:

        self.model = self._parallel_model(self.model)

    def _parallel_model(self, model):
        # TODO:统一 miner.gpu 和 miner.device 的设置
        # TODO:探索模型平行的原理，如何完成，数据集需要吗
        if self.gpu:
            gpu_count = torch.cuda.device_count()
            if gpu_count == 0:
                self.logger.warning("no GPU detected, will train on CPU.")
            else:
                self.logger.info(f"found {gpu_count} GPUs, will use all of them to train")
                devices = list(range(gpu_count))
                model.cuda()
                model = torch.nn.DataParallel(model, devices)
        return model

    def train(self):
        """
        Core Function:start to train the model
        :return:
        """
        while True:
            self.current_epoch += 1
            self.plugins.call("before_train_epoch_start", epoch=self.current_epoch)
            self.model.train()  # Set Train Mode
            train_iters = len(self.train_dataloader)  # For future change during training

            total_train_loss = 0
            self.logger.info(f"start to train epoch {self.current_epoch}")
            t = self.tqdm(self.train_dataloader)
            for index, data in enumerate(t):
                self.plugins.call(
                    "before_train_iteration_start",
                    data=data,
                    index=index,
                    total_iters=train_iters,
                    iteration=self.current_train_iteration,
                )
                train_loss = self._run_train_iteration(data)
                self.plugins.call(
                    "after_train_iteration_end",
                    loss=train_loss,
                    data=data,
                    index=index,
                    total_iters=train_iters,
                    iteration=self.current_train_iteration,
                )
                t.set_postfix({"train loss": train_loss})
                if int((index + 1) % self.accumulated_iter) == 0:
                    if self.amp and self.amp_scaler:
                        self.scaler.step(self.optimizer)
                        self.scaler.update()
                    else:
                        self.optimizer.step()
                    if self.amp and self.amp_scaler:
                        self.optimizer.zero_grad()
                    else:
                        self.optimizer.zero_grad(set_to_none=True)
                total_train_loss += train_loss
            # DataLoader End
            if self.amp and self.amp_scaler:
                self.optimizer.zero_grad()
            else:
                self.optimizer.zero_grad(set_to_none=True)

            total_train_loss = total_train_loss / train_iters
            self.logger.info(
                f"training of epoch {self.current_epoch} finished, "
                f"loss is {total_train_loss}"
            )

            self.plugins.call(
                "after_train_epoch_end",
                train_loss=total_train_loss,
                epoch=self.current_epoch,
            )

            # Begin eval
            if not self.current_epoch % self.eval_epoch and self.val_dataloader:
                self.plugins.call(
                    "before_val_epoch_start",
                    epoch=self.current_epoch,
                )
                self.model.eval()
                total_val_loss = 0
                val_iters = len(self.val_dataloader)
                with torch.no_grad():
                    self.logger.info(f"validate epoch {self.current_epoch}")
                    t = self.tqdm(self.val_dataloader)
                    for index, data in enumerate(t):
                        self.plugins.call(
                            "before_val_iteration_start",
                            data=data,
                            index=index,
                            val_iters=val_iters,
                            iteration=self.current_val_iteration,
                        )
                        predict, val_loss = self._run_val_iteration(data)
                        self.plugins.call(
                            "after_val_iteration_ended",
                            predicts=predict,
                            loss=val_loss,
                            data=data,
                            index=index,
                            val_iters=val_iters,
                            iteration=self.current_val_iteration,
                        )
                        t.set_postfix({"val loss": val_loss})
                        total_val_loss += val_loss

                total_val_loss = total_val_loss / val_iters
                self.logger.info(
                    f"validation of epoch {self.current_epoch} "
                    f"finished, loss is {total_val_loss}"
                )
                # persist model
                if total_val_loss < self.lowest_val_loss:
                    self.logger.info(
                        f"current val loss {total_val_loss} is lower than lowest {self.lowest_val_loss}, "
                        f"persist this model as best one"
                    )
                    self.lowest_val_loss = total_val_loss
                    self.persist("best")

                self.plugins.call(
                    "after_val_epoch_end",
                    val_loss=total_train_loss,
                    epoch=self.current_epoch,
                )
            self.persist("latest")
            # if self.drawer is not None:
            #     png_file = self.drawer.scalars(
            #         self.current_epoch,
            #         {"train": total_train_loss, "val": total_val_loss},
            #         "loss",
            #     )
            #     if png_file is not None:
            #         self.update_sheet(
            #             "loss", {"raw": png_file, "processor": "upload_image"}
            #         )

            if total_train_loss < self.lowest_train_loss:
                self.lowest_train_loss = total_train_loss

            if not self.current_epoch % self.persist_stride:
                self.persist("epoch_{}".format(self.current_epoch))

            if self.current_epoch >= self.max_epochs:
                self.plugins.call("before_quit")
                self.logger.warning("exceed max epochs, quit!")
                break

    def _run_train_iteration(self, data):
        self.status = "train"  # TODO:self.status Unused
        self.current_train_iteration += 1
        if self.amp and self.amp_scaler:
            with torch.cuda.amp.autocast():
                _, loss = self.forward(data)
                separate_loss = loss / self.accumulated_iter
            separate_loss = self.scaler.scale(separate_loss)
        else:
            _, loss = self.forward(data)
            separate_loss = loss / self.accumulated_iter  # TODO:实现accumulated_iter
        separate_loss.backward()
        loss = loss.detach().cpu().item()
        return loss

    def _run_val_iteration(self, data):
        self.status = "val"
        self.current_val_iteration += 1
        predict, loss = self.forward(data)
        loss = loss.detach().cpu().item()
        return predict, loss

    def forward(self, data):
        """
        A Function to calculate Network Forward results.
        The custom Forward_fn should return Network Output and Loss together.
        If Error Occurs in this Phase, Please use custom forward function
        :param data:
        :return:
        """
        predict = self.model(data.to(self.devices))
        loss = self.loss_func(predict, data.to(self.devices))
        return predict, loss

    def persist(self, name):
        """save the model to disk"""
        self.plugins.call("before_checkpoint_persisted", checkpoint_name=name)

        if isinstance(self.model, torch.nn.DataParallel):
            model_state_dict = self.model.module.state_dict()
        else:
            model_state_dict = self.model.state_dict()

        state = {
            "state_dict": model_state_dict,
            "optimizer": self.optimizer.state_dict(),
            "epoch": self.current_epoch,
            "train_iteration": self.current_train_iteration,
            "val_iteration": self.current_val_iteration,
            "lowest_train_loss": self.lowest_train_loss,
            "lowest_val_loss": self.lowest_val_loss,
        }

        state.update(self.plugins.save())

        if self.amp and self.amp_scaler:
            state["scaler"] = self.scaler.state_dict()

        modelpath = self._standard_model_path(name)
        torch.save(state, modelpath)
        self.logger.info(f"save checkpoint to {self._standard_model_path(name)}")
        self.plugins.call("after_checkpoint_persisted", modelpath=modelpath, checkpoint_name=name)

    def _standard_model_path(self, model_name):
        return self.models_dir / f"{model_name}.pth.tar"

    def _search_model_file(self, model_name):
        model_name_path = Path(str(model_name))
        models_dir_path = Path(self.models_dir)

        search_paths = [
            model_name_path,
            models_dir_path / model_name_path,
            models_dir_path / f"{model_name}.pth.tar",
            models_dir_path / f"epoch_{model_name}.pth.tar",
        ]

        for path in search_paths:
            if path.is_file():
                return path.resolve()

        return None

    # # TODO: implement methods below
    # def graceful_stop(self):
    #     """stop train and exist after this epoch"""
    #     pass
    #
    # def save_and_stop(self):
    #     """save the model immediately and stop training"""
    #     pass

    def _create_dirs(self):
        """Create directories"""
        if self.alchemy_directory:
            utils.create_dir(self.alchemy_directory)
            utils.create_dir(self.alchemy_directory, self.experiment)
            utils.create_dir(self.alchemy_directory, self.experiment, "models")



from TorchMiner import BasePlugin


class GarbageCollect(BasePlugin):
    def __init__(self):
        import gc
        self.gc = gc
        super(GarbageCollect, self).__init__()

    def after_train_epoch_end(self, *args, **kwargs):
        self._collect()

    def after_val_epoch_end(self, *args, **kwargs):
        self._collect()

    def _collect(self):
        self.logger.debug("Garbage Collected!")
        self.gc.collect()



# -*- coding:utf-8 -*-
import time

import tqdm
from IPython.core.display import HTML, display

from TorchMiner import BasePlugin


class JupyterLogger(BasePlugin):
    config = {
        "debug": ["〰", "#b2c3db"],
        "info": ["💬", "#6f818a"],
        "success": ["✅", "#7cb305"],  # TODO:How to implement success method
        "error": ["❌", "#f3715c"],
        "critical": ["❗", "#cf1322"],
        "warning": ["⚠️", "#d46b08"],
    }

    def __init__(self, name="", hide_name=True):  # name is set here to be compatible with DefaultLogger
        self.name = name
        self.hide_name = hide_name
        super(JupyterLogger, self).__init__()

    def prepare(self, miner, *args, **kwargs):
        super(JupyterLogger, self).prepare(miner)
        if not self.miner.in_notebook:
            self.logger.critical("Miner.in_notebook was set to False, but JupyterLogger will still patch on it.")
        miner.get_logger = self.get_logger

    @staticmethod
    def get_logger(name=""):
        return JupyterLogger(name)

    def _output(self, message, mode: str):
        # TODO Add More Fancy things in it, such as bold fonts and underlines or highlights
        if self.hide_name:
            display(
                HTML(
                    f'<div style="font-size: 12px; color: {self.config[mode][1]}">'
                    f'⏰ {time.strftime("%b %d - %H:%M:%S")} >>> '
                    f"{self.config[mode][0]} {message}"
                    "</div>"
                )
            )
        else:
            display(
                HTML(
                    f'<div style="font-size: 12px; color: {self.config[mode][1]}">'
                    f'⏰ {time.strftime("%b %d - %H:%M:%S")} >>> '
                    f"{self.config[mode][0]} [{self.name}] {message}"
                    "</div>"
                )
            )

    def debug(self, message):
        self._output(message, "debug")

    def info(self, message, **kwargs):  # TODO:Waiting for a better way to achieve Downward compatibility
        if "persist this model as best one" in message:
            self._output(message, "success")
        else:
            self._output(message, "info")

    def warning(self, message):
        self._output(message, "warning")

    def error(self, message):
        self._output(message, "error")

    def critical(self, message):
        self._output(message, "critical")

    def before_train_epoch_start(self, *args, **kwargs):
        display(
            HTML(
                '<div style="display: flex; justify-content: center;">'
                f'<h3 style="color: #7cb305; border-bottom: 4px dashed #91d5ff; padding-bottom: 6px;">Epoch {kwargs["epoch"]}</h3>'
                "</div>"
            )
        )


class JupyterTqdm(BasePlugin):
    def before_init(self, *args, **kwargs):
        super(JupyterTqdm, self).before_init()
        if not self.miner.in_notebook:
            self.logger.critical("Miner.in_notebook was set to False, but JupyterTqdm will still patch on it.")
        self.miner.tqdm = tqdm.notebook.tqdm



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['torchminer',
 'torchminer.plugins',
 'torchminer.plugins.Logger',
 'torchminer.plugins.Metrics',
 'torchminer.plugins.Recorder']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=8.4.0,<9.0.0',
 'google-api-python-client>=2.31.0,<3.0.0',
 'ipython>=7.18.0,<8.0.0',
 'matplotlib>=3.5.0,<4.0.0',
 'pandas>=1.3.4,<2.0.0',
 'pylint>=2.12.1,<3.0.0',
 'seaborn>=0.11.2,<0.12.0',
 'sklearn>=0.0,<0.1',
 'tensorboardX>=2.4.1,<3.0.0',
 'torch>=1.8.0,<2.0.0',
 'tqdm>=4.50.0,<5.0.0']

setup_kwargs = {
    'name': 'torchminer',
    'version': '0.4.7',
    'description': 'Run Torch With A Simple Miner',
    'long_description': 'Published on [pypi](https://pypi.org/project/torchminer/)\n\nPackaged Using [Poetry](https://python-poetry.org/)\n\n# Description\nTorchMiner is designed to automatic process the training ,evaluating and testing process for PyTorch DeepLearning,with a simple API.\n\nYou can access all Functions of MineTorch simply use `Miner`.\n\n## Quick Start\n\n```python\nimport TorchMiner\nfrom TorchMiner import Miner\nfrom TorchMiner.plugins.Logger.Jupyter import JupyterLogger, JupyterTqdm\nfrom TorchMiner.plugins.Metrics import MultiClassesClassificationMetric\nfrom TorchMiner.plugins.Recorder import TensorboardDrawer\n\nminer = Miner(\n    alchemistic_directory=\'/the/route/to/log\', \n    train_dataloader=train_dataloader, \n    val_dataloader=val_dataloader,  \n\n    model=model, \n    loss_func=MSELoss,  \n    optimizer=optimizer,  \n    experiment="the-name-of-experiment",  # Subdistribution in the experimental directory\n    resume=True,  # Whether to automatically load the previous model\n    eval_epoch=1,  # How many rounds are evaluated\n    persist_epoch=2,  # How many rounds are saved once a checkpoint\n    accumulated_iter=1,  # How many times iterates the parameter update after accumulation\n    in_notebook=True,\n    amp=True,  # Whether to use amp\n    plugins=[\n        # Use the plugins to extend the function of miner\n        JupyterLogger(),\n        JupyterTqdm(),\n        # The two above plugins are designed to get better output in Jupyter Enviroment\n        MultiClassesClassificationMetric(),\n        # This Plugin can automaticly calculate Accuracy, kappa score and Confusion Matrix in Classification problems.\n        TensorboardDrawer(input_to_model),\n        # This Plugin can record the informations generate by training process or by other plugins in Tensorboard.\n    ],\n)\n\n# And then, trigger the training process by\nminer.train()',
    'author': 'InEase',
    'author_email': 'inease28@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<4.0',
}


setup(**setup_kwargs)



# -*- coding:utf-8 -*-
from tensorboardX import SummaryWriter

from TorchMiner.plugins import BasePlugin


class TensorboardDrawer(BasePlugin):
    """To visualize everything in training process using tensorboard"""

    def __init__(self, input_to_model=None):
        super(TensorboardDrawer, self).__init__()
        self.input_to_model = input_to_model

    def prepare(self, miner, *args, **kwargs):
        super(TensorboardDrawer, self).prepare(miner)
        self.writer = SummaryWriter(log_dir=self.miner.experiment_dir)

    def after_init(self, **ignore):
        if self.input_to_model:
            try:
                self.writer.add_graph(self.miner.model, self.input_to_model)
            except Exception as e:
                self.logger.error(f"{e} occurred when visializing model")

    def after_train_epoch_end(self, train_loss, epoch, **ignore):
        self.writer.add_scalar("Loss/train", train_loss, global_step=epoch)

    def after_val_epoch_end(self, val_loss, epoch, **ignore):
        self.writer.add_scalar("Loss/val", val_loss, global_step=epoch)

    # --- Outer APIs ---
    def scalar(self, label, value, epoch=None, **ignore):
        if not epoch:
            epoch = self.miner.current_epoch
        self.writer.add_scalar(label, value, epoch)

    def figure(self, label, value, epoch=None, **ignore):
        if not epoch:
            epoch = self.miner.current_epoch
        self.writer.add_image(label, value, epoch)



# -*- coding:utf-8 -*-
import functools
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime

from TorchMiner.plugins import BasePlugin

pool = ThreadPoolExecutor(1)


# TODO:Find a better way to accomplish async function
def _async(fn):
    @functools.wraps(fn)
    def _func(*args, **kwargs):
        def _inner(*args, **kwargs):
            try:
                fn(*args, **kwargs)
            except Exception as e:
                print(e)
                raise e

        return pool.submit(_inner, *args, **kwargs)

    return _func


class ColumnNotExistsError(Exception):
    pass


class BaseSheet(BasePlugin):
    def __init__(self):
        super().__init__()
        self.columns = []
        self.cached_row_data = {}
        self.last_flushed_at = None

    def _create_experiment_row(self):
        """Create a row for this experiment"""
        raise NotImplementedError

    def _create_end_column_divider(self):
        raise NotImplementedError

    @_async
    def update(self, key, value):
        """Update value for a column"""
        if key not in self.columns:
            raise ColumnNotExistsError
        if not isinstance(value, dict):
            value = {"raw": value}
        self.cached_row_data[key] = value

    def flush(self):
        raise NotImplementedError

    def periodly_flush(self, force=False):
        now = int(datetime.now().timestamp())
        # flush every 10 seconds
        if not force and now - self.last_flushed_at < 10:
            return
        self.flush()
        self.last_flushed_at = now

    def onready(self):
        """Called after all the columns are created"""
        raise NotImplementedError

    def create_column(self, key, title):
        """Create a column on the sheet."""
        self.columns.append(key)

    @property
    def experiment_row_name(self):
        return f"_row_{self.code}"

    @property
    def title_row_name(self):
        return "_row_title"

    @property
    def end_column_name(self):
        return "_column_end_divider"

    @property
    def banner_row_name(self):
        return "_row_banner"
